#include <stdio.h>
#include <string.h>
#include "asmtypes.h"

extern "C" {
int lookup_tbl( char **tbl, char *it );
int yyparse();
}

char *reg8_tbl[] = {
  "al", "cl", "dl", "bl", "ah", "ch", "dh", "bh", 0
};

char *reg16_tbl[] = {
  "ax", "cx", "dx", "bx", "sp", "bp", "si", "di", 0
};

char *reg32_tbl[] = {
  "eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi", 0
};

int lookup_tbl( char **tbl, char *it ) {
  int i;

  for ( i = 0; tbl[i]; i++ ) {
	if ( !strcmp( tbl[i], it ) ) return i;
  }
  return -1;
}

bool code32_seg = true;

#define REG_AX_CODE 0
#define REG_DX_CODE 2
#define REG_BX_CODE 3
#define REG_BP_CODE 5
#define REG_SI_CODE 6
#define REG_DI_CODE 7

static bool fix_baseidx( operand_struct *op ) {
  if ( op->base_reg != -1 && op->base_reg != REG_BX_CODE &&
  		op->base_reg != REG_BP_CODE ) {
	if ( op->idx_reg == REG_BX_CODE || op->idx_reg == REG_BP_CODE ||
					op->idx_reg == -1 ) {
		int tmp = op->idx_reg;
		op->idx_reg = op->base_reg;
		op->base_reg = tmp;
	}
	else {
		puts( "invalid memory reference" );
		return false;
	}
  }
  if ( op->idx_reg != REG_SI_CODE && op->idx_reg != REG_DI_CODE &&
				op->idx_reg != -1 ) {
	puts( "invalid memory reference" );
	return false;
  }

  return true;
}

static int compute_modrm32( operand_struct *regmem, operand_struct *reg ) {
  int modrm = 0, disp;

  if ( regmem->have_disp ) {
	disp = regmem->disp;
	if ( disp >= -0x7f && disp <= 0x7f ) {
		regmem->dispw = 8;
		modrm |= 0x40;
	}
	else {
       		regmem->dispw = 32;
		modrm |= 0x80;
        }
  }
  else {
	if ( regmem->base_reg == REG_BP_CODE ) {
		regmem->have_disp = true;
		regmem->dispw = 8;
		regmem->disp = 0;
		modrm |= 0x40;
	}
  }
  if ( reg ) modrm |= reg->reg << 3 ;

  if ( regmem->base_reg != -1 && regmem->idx_reg == -1 ) {
	modrm |= regmem->base_reg;
  }
  else {
  	regmem->need_sib = 1;
	modrm |= 4;
  }

  return modrm;
}

static int compute_sib( operand_struct *regmem ) {
  int sib = 0;
  
  switch ( regmem->scale ) {
  case 1: break;
  case 2: sib |= 0x40; break;
  case 4: sib |= 0x80; break;
  case 8: sib |= 0xc0; break;
  }
  sib |= (regmem->idx_reg << 3) | regmem->base_reg;

  return sib;
}

static int compute_modrm( operand_struct *regmem, operand_struct *reg ) {
  int modrm, disp;

  if ( !fix_baseidx( regmem ) ) return 0;
 
  modrm = 0; 
  switch ( regmem->base_reg ) {
  case REG_BX_CODE:
	 break;
  case REG_BP_CODE:
	 modrm |= 2;
	 break;
  case -1:
	 modrm |= 4;
	 break;
  }
  switch ( regmem->idx_reg ) {
  case REG_SI_CODE:
	break;
  case REG_DI_CODE:
	modrm |= 1;
	break;
  case -1:
	if ( regmem->base_reg == REG_BX_CODE ) modrm = 7;
	else if ( regmem->base_reg == REG_BP_CODE ) {
		if ( !regmem->have_disp ) {
			regmem->have_disp = 1;
			regmem->disp = 0;
		}
	}
	else modrm = 6;
	break;
  }

  if ( regmem->have_disp ) {
	disp = regmem->disp;
	if ( disp >= -0x7f && disp <= 0x7f ) {
		regmem->dispw = 8;
		modrm |= 0x40;
	}
        else if ( disp >= -0x7fff && disp <= 0x7fff ) {
		regmem->dispw = 16;
		modrm |= 0x80;
	}
	else {
        	regmem->dispw = 32;
		modrm |= 0x80;
        }
  }

  if ( reg ) modrm |= reg->reg << 3;

  return modrm;
}

unsigned char asmbuf_x86[50], *asmbuf_ptr_x86 = asmbuf_x86;

static void put_byte( int val ) {
  *asmbuf_ptr_x86++ = val;
}

static void put_imm( int imm, int width ) {
  if ( width == 8 ) put_byte( imm );
  else if ( width == 16 ) {
	put_byte( imm & 0xff );
	put_byte( imm >> 8 );
  }
  else if ( width == 32 ) {
	put_byte( imm & 0xff );
	put_byte( (imm >> 8) & 0xff );
	put_byte( (imm >> 16) & 0xff );
	put_byte( (imm >> 24) & 0xff );
  }
}

static void put_csize_prefix( int opsize ) {
  if ( opsize == 32 && !code32_seg || opsize == 16 && code32_seg )
	put_byte( 0x66 );
}

static void put_modregrm( operand_struct *regrm, int reg_code ) {
  if ( code32_seg ) {
	put_byte( compute_modrm32( regrm, 0 ) | reg_code );
	if ( regrm->need_sib ) put_byte( compute_sib( regrm ) );
  }
  else
	put_byte( compute_modrm( regrm, 0 ) | reg_code );
  if ( regrm->have_disp ) put_imm( regrm->disp, regrm->dispw );
}

static void put_modregrm( operand_struct *regrm, operand_struct *reg ) {
  put_modregrm( regrm, reg->reg << 3 );
}

struct cmdtbl_ent {
  char *mnem;
  int argnum;
  void (*cmd_proc)( cmdtbl_ent *ent, cmd_struct *cmd );
};

static void aaa_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0x37 );
}

static void aad_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0xd5 );
  put_byte( 0x0a );
}

static void aam_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0xd4 );
  put_byte( 0x0a );
}

static void aas_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0x3f );
}

static void alu_group_proc( cmd_struct *cmd, int code, int ext1, int ext2 ) {
  operand_struct r1, r2;
  int modrm, sib;

  r1 = cmd->op[0];
  r2 = cmd->op[1];

  if ( r1.type == OP_REG && r2.type == OP_REG ) {
        if ( r1.width != r2.width ) {
                puts( "operands size mismatch" );
                return;
        }

        if ( r1.width == 8 ) put_byte( code ); else put_byte( code | 1 );
        put_byte( 0xc0 | (r2.reg << 3) | r1.reg );
  }
  else if ( r1.type == OP_REG && r2.type == OP_MEM ) {
        if ( r1.width == 8 ) put_byte( code | 2 ); else put_byte( code | 3 );
	put_modregrm( &r2, &r1 );
  }
  else if ( r1.type == OP_MEM && r2.type == OP_REG ) {
	if ( r2.width == 8 ) put_byte( code ); else put_byte( code | 1 );
	put_modregrm( &r1, &r2 );
  }
  else if ( r1.type == OP_REG && r1.reg == REG_AX_CODE && r2.type == OP_IMM ) {
	if ( r1.width == 8 ) put_byte( code | 4 ); else put_byte( code | 5 );
	put_imm( r2.val, r1.width );
  }
  else if ( r1.type == OP_REG && r2.type == OP_IMM ) {
	if ( r1.width == 8 ) put_byte( ext1 ); else put_byte( ext1 | 1 );
	put_byte( 0xc0 | ext2 | r1.reg );
	put_imm( r2.val, r1.width );
  }
  else if ( r1.type == OP_MEM && r2.type == OP_IMM ) {
//	if ( r1.width == 8 ) put_byte( 0x80 ); else put_byte( 0x81 );
	put_byte( ext1 | 1 );
	put_modregrm( &r1, ext2 );
	put_imm( r2.val, 32 );
  }
  else {
	puts( "invalid combination of operands" );
	return;
  }
}

static void adc_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  alu_group_proc( cmd, 0x10, 0x80, 0x10 );
}

static void add_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  alu_group_proc( cmd, 0, 0x80, 0 );
}

static void and_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  alu_group_proc( cmd, 0x20, 0x80, 0x20 );
}

static void or_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  alu_group_proc( cmd, 0x08, 0x80, 0x08 );
}

static void xor_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  alu_group_proc( cmd, 0x30, 0x80, 0x30 );
}

static void call_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  operand_struct r;
  int modrm;

  r = cmd->op[0];
  if ( r.type == OP_IMM ) {
	put_byte( 0xe8 );
	put_imm( r.val, 32 );
  }
  else if ( r.type == OP_REG ) {
	put_byte( 0xff );
	put_byte( 0xd0 | r.reg );
  }
  else if ( r.type == OP_MEM ) {
	put_byte( 0xff );
	put_modregrm( &r, 0x10 );
  }
}

static void jcc_group_proc( cmd_struct *cmd, int cond ) {
  if ( cmd->op[0].type != OP_IMM ) {
  	puts( "invalid operand" );
	return;
  }

  put_byte( 0x0f );
  put_byte( 0x80 | cond );
  put_imm( cmd->op[0].val, 32 );
}

static void jo_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 0 );
}

static void jno_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 1 );
}

static void jb_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 2 );
}

static void jnb_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 3 );
}

static void je_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 4 );
}

static void jne_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 5 );
}

static void jbe_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 6 );
}

static void jnbe_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 7 );
}

static void js_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 8 );
}

static void jns_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 9 );
}

static void jp_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 0xa );
}

static void jnp_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 0xb );
}

static void jl_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 0xc );
}

static void jnl_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 0xd );
}

static void jle_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 0xe );
}

static void jnle_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  jcc_group_proc( cmd, 0xf );
}

static void lea_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  operand_struct r1, r2;
  int modrm;

  r1 = cmd->op[0];
  r2 = cmd->op[1];
  if ( r1.type == OP_REG && r2.type == OP_MEM ) {
	put_byte( 0x8d );
	put_modregrm( &r2, &r1 );
  }
  else
	puts( "invalid combination of operands" );
}

static void leave_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0xc9 );
}

static void pushad_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0x60 );
}

static void popad_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0x61 );
}

static void mov_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  operand_struct r1, r2;

  r1 = cmd->op[0];
  r2 = cmd->op[1];
  if ( r1.type == OP_REG && r2.type == OP_IMM ) {
	if ( r1.width == 8 ) put_byte( 0xb0 | r1.reg );
	else {
		put_csize_prefix( r1.width );
		put_byte( 0xb8 | r1.reg );
	}
	put_imm( r2.val, r1.width );
  }
  else if ( r1.type == OP_REG && r1.reg == REG_AX_CODE &&
	    r2.type == OP_MEM && r2.base_reg == -1 && r2.idx_reg == -1 ) {
	if ( r1.width == 8 ) put_byte( 0xa0 );
	else {
		put_csize_prefix( r1.width );
		put_byte( 0xa1 );
	}
	put_imm( r2.disp, 32 );
  }
  else if ( r1.type == OP_MEM && r1.base_reg == -1 && r1.idx_reg == -1 &&
	    r2.type == OP_REG && r2.base_reg == REG_AX_CODE ) {
	if ( r2.width == 8 ) put_byte( 0xa0 );
	else {
		put_csize_prefix( r2.width );
		put_byte( 0xa1 );
	}
	put_imm( r1.disp, 32 );
  }
  else alu_group_proc( cmd, 0x88, 0xc6, 0 );
}

static void movx_group_proc( cmd_struct *cmd, int code ) {
  operand_struct r1, r2;
  int modrm;

  r1 = cmd->op[0];
  r2 = cmd->op[1];
  if ( r1.type == OP_REG && (r2.type == OP_REG || r2.type == OP_MEM) ) {
	put_csize_prefix( r1.width );
  	put_byte( 0x0f );
	if ( r2.type == OP_REG && r2.width == 8 )
		put_byte( code );
	else
		put_byte( code | 1 );

	if ( code32_seg ) {
		put_byte( compute_modrm32( &r2, &r1 ) );
		if ( r2.need_sib ) put_byte( compute_sib( &r2 ) );
	}
	else {
		modrm = compute_modrm( &r2, &r1 );
        	put_byte( modrm );
	}
        if ( r2.have_disp ) put_imm( r2.disp, r2.dispw );
  }
}

static void movzx_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  movx_group_proc( cmd, 0xb6 );
}

static void movsx_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  movx_group_proc( cmd, 0xbe );
}

static void stack_group_proc( cmd_struct *cmd, int code, int ext1, int ext2 ) {
  operand_struct r;
  int modrm;

  r = cmd->op[0];
  if ( r.type == OP_REG && (r.width == 16 || r.width == 32) ) {
	put_byte( code | r.reg );
  }
  else if ( r.type == OP_MEM ) {
	put_byte( ext1 );
	put_modregrm( &r, ext2 );
  }
  else if ( r.type == OP_IMM ) {
	put_byte( 0x68 );
	put_imm( r.val, 32 );
  }
  else {
	puts( "invalid operand" );
	return;
  }
}

static void pop_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  if ( cmd->op[0].type == OP_IMM ) {
  	puts( "invalid operand" );
	return;
  }
  stack_group_proc( cmd, 0x58, 0x8f, 0 );
}

static void push_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  stack_group_proc( cmd, 0x50, 0xff, 0x30 );
}

static void ret_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0xc3 );
}

static void sbb_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  alu_group_proc( cmd, 0x18, 0x80, 0x18 );
}

static void sub_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  alu_group_proc( cmd, 0x28, 0x80, 0x28 );
}

static void cmp_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  alu_group_proc( cmd, 0x38, 0x80, 0x38 );
}

static void test_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  operand_struct r1, r2;

  r1 = cmd->op[0];
  r2 = cmd->op[1];
  if ( r1.type == OP_REG && r2.type == OP_REG && r1.width == r2.width ) {
  	if ( r1.width == 8) put_byte( 0x84 );
	else {
		put_csize_prefix( r1.width );
		put_byte( 0x85 );
	}
	put_byte( 0xc0 | (r1.reg << 3) | r2.reg );
  }
  else if ( r1.type == OP_REG && r2.type == OP_MEM ||
  	    r1.type == OP_MEM && r2.type == OP_REG ) {
	
	if ( r1.type == OP_MEM ) {
		operand_struct tmp;
		tmp = r1;
		r1 = r2;
		r2 = tmp;
	}
  	if ( r1.width == 8 ) put_byte( 0x84 );
	else {
		put_csize_prefix( r1.width );
		put_byte( 0x85 );
	}
	put_modregrm( &r2, &r1 );
  }
  else if ( r1.type == OP_REG && r1.reg == REG_AX_CODE && r2.type == OP_IMM ) {
  	if ( r1.width == 8 ) put_byte( 0xa8 );
	else {
		put_csize_prefix( r1.width );
		put_byte( 0xa9 );
	}
	put_imm( r2.val, r1.width );
  }
  else if ( r1.type == OP_REG && r2.type == OP_IMM ) {
  	if ( r1.width == 8 ) put_byte( 0xf6 );
	else {
		put_csize_prefix( r1.width );
		put_byte( 0xf7 );
	}
	put_byte( 0xc0 | r1.reg );
	put_imm( r2.val, r1.width );
  }
  else if ( r1.type == OP_MEM && r2.type == OP_IMM ) {
  	if ( r2.width == 8 ) put_byte( 0xf6 );
	else
		put_byte( 0xf7 );
	put_modregrm( &r1, 0 );
	put_imm( r2.val, 32 );
  }
}

static void xchg_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  operand_struct r1, r2;

  r1 = cmd->op[0];
  r2 = cmd->op[1];
  if ( r1.type == OP_REG && r2.type == OP_REG ) {
  	if ( r1.reg == REG_AX_CODE )
	  	put_byte( 0x90 | r2.reg );
	else if ( r2.reg == REG_AX_CODE )
		put_byte( 0x90 | r1.reg );
  }
  else if ( r1.type == OP_REG && r2.type == OP_MEM ||
            r1.type == OP_MEM && r2.type == OP_REG ) {

	if ( r1.type == OP_MEM ) {
		operand_struct tmp;
		tmp = r1;
		r1 = r2;
		r2 = tmp;
	}
  	if ( r1.width == 8 ) put_byte( 0x86 );
	else {
		put_csize_prefix( r1.width );
		put_byte( 0x87 );
	}
	put_modregrm( &r2, &r1 );
  }
}

static void xlat_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_byte( 0xd7 );
}

static void cbw_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_csize_prefix( 16 );
  put_byte( 0x98 );
}

static void cwd_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_csize_prefix( 16 );
  put_byte( 0x99 );
}

static void cwde_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_csize_prefix( 32 );
  put_byte( 0x98 );
}

static void cdq_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  put_csize_prefix( 32 );
  put_byte( 0x99 );
}

static void inout_group_proc( int code1, int code2,
	operand_struct& r1, operand_struct& r2 ) {

  if ( r1.type == OP_REG && r1.reg == REG_AX_CODE ) {
  	if ( r2.type == OP_IMM ) {
		if ( r1.width == 8 ) put_byte( code1 );
		else put_byte( code1 | 1 );
		put_imm( r2.val, 8 );
	}
	else if ( r2.type == OP_REG && r2.reg == REG_DX_CODE ) {
		if ( r1.width == 8 ) put_byte( code2 );
		else put_byte( code2 | 1 );
	}
  }
}

static void in_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  inout_group_proc( 0xe4, 0xec, cmd->op[0], cmd->op[1] );
}

static void out_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  inout_group_proc( 0xe6, 0xee, cmd->op[1], cmd->op[0] );
}

static void oneop_alu_group_proc( cmd_struct *cmd, int code2, int ext ) {
  operand_struct r1;

  r1 = cmd->op[0];
  	
  if ( r1.width == 8 ) put_byte( code2 );
  else {
	put_csize_prefix( r1.width );
	put_byte( code2 | 1 );
  }
  put_modregrm( &r1, ext );
}

static void inc_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  operand_struct r1 = cmd->op[0];
  if ( r1.type == OP_REG )
  	put_byte( 0x40 | r1.reg );
  else
  	oneop_alu_group_proc( cmd, 0xfe, 0 );
}

static void dec_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  operand_struct r1 = cmd->op[0];
  if ( r1.type == OP_REG )
 	put_byte( 0x48 | r1.reg );
  else
	oneop_alu_group_proc( cmd, 0xfe, 0x08 );
}

static void neg_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  oneop_alu_group_proc( cmd, 0xf6, 0x18 );
}

static void not_proc( cmdtbl_ent *ent, cmd_struct *cmd ) {
  oneop_alu_group_proc( cmd, 0xf6, 0x10 );
}

cmdtbl_ent cmd_table[] = {
  { "aaa", 0, aaa_proc },
  { "aad", 0, aad_proc },
  { "aam", 0, aam_proc },
  { "aas", 0, aas_proc },
  { "adc", 2, adc_proc },
  { "add", 2, add_proc },
  { "and", 2, and_proc },
  { "call", 1, call_proc },
  { "cbw", 0, cbw_proc },
  { "cmp", 2, cmp_proc },
  { "cwd", 0, cwd_proc },
  { "cwde", 0, cwde_proc },
  { "cdq", 0, cdq_proc },
  { "dec", 1, dec_proc },
  { "in", 2, in_proc },
  { "inc", 1, inc_proc },
  { "jo", 1, jo_proc },
  { "jno", 1, jno_proc },
  { "jb", 1, jb_proc },
  { "jnb", 1, jnb_proc },
  { "je", 1, je_proc },
  { "jne", 1, jne_proc },
  { "jbe", 1, jbe_proc },
  { "jnbe", 1, jnbe_proc },
  { "js", 1, js_proc },
  { "jns", 1, jns_proc },
  { "jp", 1, jp_proc },
  { "jnp", 1, jnp_proc },
  { "jl", 1, jl_proc },
  { "jnl", 1, jnl_proc },
  { "jle", 1, jle_proc },
  { "jnle", 1, jnle_proc },
  { "lea", 2, lea_proc },
  { "leave", 0, leave_proc },
  { "mov", 2, mov_proc },
  { "movsx", 2, movsx_proc },
  { "movzx", 2, movzx_proc },
  { "neg", 1, neg_proc },
  { "not", 1, not_proc },
  { "or", 2, or_proc },
  { "out", 2, out_proc },
  { "pop", 1, pop_proc },
  { "popad", 0, popad_proc },
  { "push", 1, push_proc },
  { "pushad", 0, pushad_proc },
  { "ret", 0, ret_proc },
  { "sbb", 2, sbb_proc },
  { "sub", 2, sub_proc },
  { "test", 2, test_proc },
  { "xchg", 2, xchg_proc },
  { "xlat", 0, xlat_proc },
  { "xor", 2, xor_proc },
};

static cmdtbl_ent *lookup_cmd( cmd_struct *cmd ) {
  int i, cmd_count = sizeof(cmd_table) / sizeof(cmdtbl_ent);

  if ( !cmd->mnem ) return 0;

  for ( i = 0; i < cmd_count; i++ )
        if ( !strcmp( cmd_table[i].mnem, cmd->mnem ) &&
                cmd_table[i].argnum == cmd->argnum ) return &cmd_table[i];

  return 0;
}

static void translate_cmd( cmdtbl_ent *ent, cmd_struct *cmd ) {
  ent->cmd_proc( ent, cmd );
}

cmd_struct cur_cmd;

static void dump_buf() {
  unsigned char *p = asmbuf_x86;

  while ( p != asmbuf_ptr_x86 )
        printf( "%02X ", *p++ );
  putchar( '\n' );
}

extern "C" {
void yy_scan_string( char *str );
}

bool do_asm_str_x86( char *str ) {
  yy_scan_string( str );
  if ( yyparse() ) return false;

  cmdtbl_ent *ent = lookup_cmd( &cur_cmd );
  if ( !ent ) return false;
  asmbuf_ptr_x86 = asmbuf_x86;
  translate_cmd( ent, &cur_cmd );

  return true;
}

#ifdef MAIN
extern int yydebug;

int main( int argc, char **argv ) {
  FILE *fd = stdin;
  int len;
  char str[80];

//  yydebug = 1;

  while ( !feof(fd) ) {
  	if ( !fgets( str, sizeof str, fd ) ) break;
	len = strlen(str);
	if ( str[len-1] == '\n' ) str[len-1] = 0;
	if ( do_asm_str( str ) ) dump_buf();
  }
}
#endif
