%{
#include <stdlib.h>
#include "asmtypes.h"

#define YYDEBUG 1

extern struct cmd_struct cur_cmd;
void yyerror( char *s );
%}

%union {
struct operand_struct operand;
struct cmd_struct command;
char *text;
int intval;
char charval;
}

%token <intval> reg8 reg16 reg32 int32 scale_tk
%token <text> cmd
%token <char> spc
%type <operand> arg memref16 idx32 memref32

%%
input:	/* nul input */
	| insn
;

memref16: '[' reg16 ']'	{
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = -1;
				$$.scale = 1;
				$$.have_disp = 0;
			}
	| '[' int32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = -1;
				$$.idx_reg = -1;
				$$.scale = 1;
				$$.have_disp = 1;
				$$.disp = $2;
			}
	| '[' reg16 '+' reg16 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = $4;
				$$.scale = 1;
				$$.have_disp = 0;
			}
	| '[' reg16 '+' int32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = -1;
				$$.scale = 1;
				$$.have_disp = 1;
				$$.disp = $4;
			}
	| '[' int32 '+' reg16 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $4;
				$$.idx_reg = -1;
				$$.scale = 1;
				$$.have_disp = 1;
				$$.disp = $2;
			}
	| '[' reg16 '+' reg16 '+' int32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = $4;
				$$.scale = 1;
				$$.have_disp = 1;
				$$.disp = $6;
			}
	| '[' reg16 '+' int32 '+' reg16 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = $6;
				$$.scale = 1;
				$$.have_disp = 1;
				$$.disp = $4;
			}
	| '[' int32 '+' reg16 '+' reg16 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $4;
				$$.idx_reg = $6;
				$$.scale = 1;
				$$.have_disp = 1;
				$$.disp = $2;
			}
;

memref32:	'[' reg32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = -1;
				$$.scale = 1;
				$$.have_disp = 0;
			}
	| '[' idx32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = -1;
				$$.idx_reg = $2.reg;
				$$.scale = $2.scale;
				$$.have_disp = 0;
			}
	| '[' reg32 '+' int32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = -1;
				$$.scale = 1;
				$$.have_disp = 1;
				$$.disp = $4;
			}
	| '[' int32 '+' reg32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $4;
				$$.idx_reg = -1;
				$$.scale = 1;
				$$.have_disp = 1;
				$$.disp = $2;
			}
	| '[' reg32 '+' idx32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = $4.reg;
				$$.scale = $4.scale;
				$$.have_disp = 0;
			}
	| '[' idx32 '+' reg32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $4;
				$$.idx_reg = $2.reg;
				$$.scale = $2.scale;
				$$.have_disp = 0;
			}
	| '[' idx32 '+' int32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = -1;
				$$.idx_reg = $2.reg;
				$$.scale = $2.scale;
				$$.have_disp = 1;
				$$.disp = $4;
			}
	| '[' int32 '+' idx32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = -1;
				$$.idx_reg = $4.reg;
				$$.scale = $4.scale;
				$$.have_disp = 1;
				$$.disp = $2;
			}
	| '[' reg32 '+' idx32 '+' int32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = $4.reg;
				$$.scale = $4.scale;
				$$.have_disp = 1;
				$$.disp = $6;
			}
	| '[' reg32 '+' int32 '+' idx32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $2;
				$$.idx_reg = $6.reg;
				$$.scale = $6.scale;
				$$.have_disp = 1;
				$$.disp = $4;
			}
	| '[' int32 '+' reg32 '+' idx32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $4;
				$$.idx_reg = $6.reg;
				$$.scale = $6.scale;
				$$.have_disp = 1;
				$$.disp = $2;
			}
	| '[' idx32 '+' reg32 '+' int32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $4;
				$$.idx_reg = $2.reg;
				$$.scale = $2.scale;
				$$.have_disp = 1;
				$$.disp = $6;
			}
	| '[' idx32 '+' int32 '+' reg32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $6;
				$$.idx_reg = $2.reg;
				$$.scale = $2.scale;
				$$.have_disp = 1;
				$$.disp = $4;
			}
	| '[' int32 '+' idx32 '+' reg32 ']' {
				$$.type = OP_MEM;
				$$.width = -1;
				$$.base_reg = $6;
				$$.idx_reg = $4.reg;
				$$.scale = $4.scale;
				$$.have_disp = 1;
				$$.disp = $2;
			}
;

idx32:	reg32		{
				$$.type = OP_REG;
				$$.reg = $1;
				$$.width = 32;
				$$.scale = 1;
			}
	| reg32 '*' scale_tk {
				$$.type = OP_REG;
				$$.reg = $1;
				$$.width = 32;
				$$.scale = $3;
			}
	| scale_tk '*' reg32 {
				$$.type = OP_REG;
				$$.reg = $3;
				$$.width = 32;
				$$.scale = $1;
			}
;


arg:	reg8		{
				$$.type = OP_REG;
				$$.reg = $1;
				$$.width = 8;
			}
	| reg16		{
				$$.type = OP_REG;
				$$.reg = $1;
				$$.width = 16;
			}
	| reg32		{
				$$.type = OP_REG;
				$$.reg = $1;
				$$.width = 32;
			}
	| memref16
	| memref32
	| int32		{
				$$.type = OP_IMM;
				$$.val = $1;
			}
;

insn:	cmd		{
				cur_cmd.mnem = $1;
				cur_cmd.argnum = 0;
				cur_cmd.op = 0;
			}
	| cmd arg {
				cur_cmd.mnem = $1;
				cur_cmd.argnum = 1;
				cur_cmd.op = (struct operand_struct*) malloc( sizeof(struct operand_struct) );
				cur_cmd.op[0] = $2;
			}
	| cmd arg ',' arg {
				cur_cmd.mnem = $1;
				cur_cmd.argnum = 2;
				cur_cmd.op = (struct operand_struct*) malloc( 2 * sizeof(struct operand_struct) );
				cur_cmd.op[0] = $2;
				cur_cmd.op[1] = $4;
			}
	| cmd arg ',' arg ',' arg {
				cur_cmd.mnem = $1;
				cur_cmd.argnum = 3;
				cur_cmd.op = (struct operand_struct*) malloc( 3 * sizeof(struct operand_struct) );
				cur_cmd.op[0] = $2;
				cur_cmd.op[1] = $4;
				cur_cmd.op[2] = $6;
			}
;
%%

void yyerror( char *s ) {
//  printf( "\n%s\n", s );
}

#if 0
int main() {
  yyparse();
}
#endif
