#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <locale.h>
#include <list>
#include <vector>
#include <map>
#include <string>
using namespace std;

#ifdef USE_WIDEC

#ifndef _XOPEN_SOURCE_EXTENDED
#define _XOPEN_SOURCE_EXTENDED
#endif

#include <wctype.h>
#include <cursesw.h>
const char *mb_locale  = "ru_RU.KOI8-R",
           *utf_locale = "ru_RU.UTF-8";

#else // USE_WIDEC

#include <curses.h>

#endif // USE_WIDEC

#include <libelf.h>
#include <capstone/capstone.h>

#include "disasm.h"
#include "gdbclient.h"
#include "prstatus.h"

/*
  symbol 0x9b is somehow bad, isprint returns true, but
  when we output it, cursor moves to left.
*/

struct instr {
  char mnemonic[100];
};

bool x86_mode = false, x86_64_mode = false, arm_mode = false, c166_mode = false;

struct hints {
  const char *hint[11];
};

list <hints> hints_stack;
int hint_offs[] = { 0, 1, 9, 17, 25, 33, 41, 49, 57, 65, 74 };

void set_keyhint( int n, const char *hint ) {
  attrset( COLOR_PAIR(1) );
  mvaddstr( LINES-1, hint_offs[n], hint );
  hints_stack.back().hint[n] = hint;
}

void clear_keyhints() {
  attrset( COLOR_PAIR(1) );
  for ( int i = 1; i <= 10; i++ ) {
  	hints_stack.back().hint[i] = "      ";
	mvaddstr( LINES-1, hint_offs[i], hints_stack.back().hint[i] );
  }
}

void push_keyhints() {
  hints_stack.push_back( hints_stack.back() );
}

void pop_keyhints() {
  hints_stack.pop_back();
  attrset( COLOR_PAIR(1) );
  for ( int i = 1; i <= 10; i++ )
	mvaddstr( LINES-1, hint_offs[i], hints_stack.back().hint[i] );
}

void init_keyhints() {
  hints nhints;

  attrset( COLOR_PAIR(3) );
  mvaddstr( LINES-1, 0, "1" );
  mvaddstr( LINES-1, 7, " 2" );
  mvaddstr( LINES-1, 15, " 3" );
  mvaddstr( LINES-1, 23, " 4" );
  mvaddstr( LINES-1, 31, " 5" );
  mvaddstr( LINES-1, 39, " 6" );
  mvaddstr( LINES-1, 47, " 7" );
  mvaddstr( LINES-1, 55, " 8" );
  mvaddstr( LINES-1, 63, " 9" );
  mvaddstr( LINES-1, 71, " 10" );

  hints_stack.push_back( nhints );
  clear_keyhints();
}

list <WINDOW*> win_stack;

void pop_winstack() {
  win_stack.pop_back();

  touchwin( stdscr );
  refresh();

  list<WINDOW*>::iterator w = win_stack.begin();
  for ( ; w != win_stack.end(); ++w ) {
  	touchwin( *w );
  	wrefresh( *w );
  }
}

struct string_ent_state {
  int y, x, idx;
  bool hi_nibble;
  unsigned char *buf;
  int bsz;
  WINDOW *win;
};

void init_cstr_entry( string_ent_state *st, WINDOW *win,
	unsigned char *buf, int bsz ) {
  st->win = win;
  st->buf = buf;
  st->bsz = bsz;
  
  getyx( st->win, st->y, st->x );
  whline( st->win, ACS_CKBOARD, st->bsz - 1 );
  wmove( st->win, st->y, st->x );
  st->idx = 0;
}

void init_xstr_entry( string_ent_state *st, WINDOW *win,
	unsigned char *buf, int bsz ) {
  st->win = win;
  st->buf = buf;
  st->bsz = bsz;
  
  getyx( st->win, st->y, st->x );
  whline( st->win, ACS_CKBOARD, st->bsz * 3 - 1 );
  wmove( st->win, st->y, st->x );
  st->idx = 0;
  st->hi_nibble = true;
}

int enter_cstring( string_ent_state *st, int& c, bool single_char ) {
  wmove( st->win, st->y, st->x + st->idx );
  for (;;) {
  	c = wgetch( st->win );
	if ( c == '\n' ) return 1;
	else if ( c == 033 ) return 2;
	else if ( c == '\t' ) return 0;
	else if ( c == (c & 0xff) && st->idx < st->bsz ) {
		if ( c != 0x9b && isprint( c ) ) waddch( st->win, c );
		else waddch( st->win, '.' );
		st->buf[st->idx++] = c;
	}
	else if ( c == KEY_BACKSPACE && st->idx ) {
		st->idx--;
		mvwaddch( st->win, st->y, st->x + st->idx, ACS_CKBOARD );
		wmove( st->win, st->y, st->x + st->idx );
	}
	else break;
	if ( single_char ) break;
  }

  return 0;
}

void update_cstring( string_ent_state *st ) {
  int i, c;

  wmove( st->win, st->y, st->x );
  for ( i = 0; i < st->idx; i++ ) {
  	c = st->buf[i];
	if ( c == 0x9b || !isprint( c ) ) c = '.';
	waddch( st->win, c );
  }
  for (; i < st->bsz; i++ ) waddch( st->win, ACS_CKBOARD );
  wrefresh( st->win );
}

int enter_xstring( string_ent_state *st, int& c, bool single_char ) {
  int n, xpos;
  char s[2];

  if ( st->hi_nibble ) {
  	if ( st->idx ) xpos = st->x + st->idx * 3 - 1;
	else xpos = st->x;
  }
  else xpos = st->x + st->idx * 3 + 1;
  wmove( st->win, st->y, xpos );

  for (;;) {
  	c = wgetch( st->win );
	if ( c == (c & 0xff) && isxdigit(c) && st->idx < st->bsz ) {
		s[0] = c;
		s[1] = 0;
		n = strtol( s, 0, 16 );
		if ( st->hi_nibble ) {
			if ( st->idx ) waddch( st->win, ' ' );
			waddch( st->win, toupper(c) );
			st->buf[st->idx] = n << 4;
		}
		else {
			waddch( st->win, toupper(c) );
			n |= st->buf[st->idx] & 0xf0;
			st->buf[st->idx++] = n;
		}
		st->hi_nibble = !st->hi_nibble;
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( st->hi_nibble ) {
			if ( !st->idx ) continue;
			mvwaddch( st->win, st->y, st->x + st->idx * 3 - 2,
				ACS_CKBOARD );
			wmove( st->win, st->y, st->x + st->idx * 3 - 2 );
			st->idx--;
			st->buf[st->idx] &= 0xf0;
		}
		else {
			if ( st->idx ) {
				mvwaddch( st->win, st->y,
					st->x + st->idx * 3 - 1, ACS_CKBOARD );
				waddch( st->win, ACS_CKBOARD );
				wmove( st->win, st->y,
					st->x + st->idx * 3 - 1 );
			}
			else {
				mvwaddch( st->win, st->y, st->x, ACS_CKBOARD );
				wmove( st->win, st->y, st->x );
			}
		}
		st->hi_nibble = !st->hi_nibble;
	}
	else if ( st->hi_nibble && c == '\n' ) return 1;
	else if ( c == 033 ) return 2;
	else break;
	if ( single_char ) break;
  }
  return 0;
}

void update_xstring( string_ent_state *st ) {
  int i, c;

  wmove( st->win, st->y, st->x );
  for ( i = 0; i < st->idx; i++ ) {
  	c = st->buf[i];
	if ( i ) waddch( st->win, ' ' );
	wprintw( st->win, "%02X", c );
  }
  for (; i < st->bsz; i++ ) {
  	waddch( st->win, ACS_CKBOARD );
  	waddch( st->win, ACS_CKBOARD );
  	if ( i != st->bsz-1 ) waddch( st->win, ACS_CKBOARD );
  }
  wrefresh( st->win );
}
void load_hiew_vmm();

char *fname;

void disasm_init( char *fname0 ) {
  int i;

  fname = strdup( fname0 );

  init_pair( 1, COLOR_BLACK, COLOR_CYAN );
  init_pair( 2, COLOR_CYAN, COLOR_BLUE );
  init_pair( 3, COLOR_WHITE, COLOR_BLACK );
  init_pair( 4, COLOR_BLACK, COLOR_WHITE );
  init_pair( 5, COLOR_YELLOW, COLOR_BLUE );
  init_pair( 6, COLOR_BLUE, COLOR_CYAN );
  move( 0, 0 );
  attrset( COLOR_PAIR(1) );
  for ( i = 0; i < COLS; i++ ) addch(' ');
  attrset( COLOR_PAIR(2) );
  for ( i = 0; i < (LINES-2)*COLS; i++ ) addch( ' ' );
  attrset( COLOR_PAIR(1) );
  for ( i = 0; i < COLS; i++ ) addch(' ');
  mvaddstr( 0, 1, fname );
  addch( ' ' );
  addch( ACS_VLINE );
  init_keyhints();
  set_keyhint( 1, "About" );
  set_keyhint( 3, "Edit" );
  set_keyhint( 4, "Mode" );
  set_keyhint( 5, "Goto" );
  set_keyhint( 6, "ObjTbl" );
  set_keyhint( 7, "Search" );
  set_keyhint( 10, "Quit" );
//  load_hiew_vmm();
}

char buf[512];
int fd, buf_offs, fsize;
bool gdb_connection = false;

unsigned char *get_data( int offs ) {
  if ( offs < 0 || offs >= fsize ) return 0;
  if ( offs < buf_offs || offs > (int)(buf_offs + sizeof(buf) - 40) ) {
  	if ( offs - (int) sizeof(buf)/2 > 0 ) buf_offs = offs - sizeof(buf)/2;
	else buf_offs = 0;
	
	if ( gdb_connection )
		gdb_readmem( (unsigned char*)buf, buf_offs, sizeof buf );
	else {
	  	lseek( fd, buf_offs, SEEK_SET );
		read( fd, buf, sizeof buf );
	}
  }
  return (unsigned char*)buf + offs - buf_offs;
}

struct serv_desc_entry {
  int serv_num;
  char *serv_desc;
  vector <char*> func;
};

vector <serv_desc_entry> serv_list;

void load_hiew_vmm() {
  serv_desc_entry desc;
  char buf[80], *p, *serv_desc;
  FILE *f;
  int serv_num, rc;

  f = fopen( "/usr/local/lib/disasm/hiew.vmm", "r" );
  if ( !f ) {
  	f = fopen( "hiew.vmm", "r" );
	if ( !f ) return;
  }
  while ( !feof( f ) ) {
  	if ( !fgets( buf, sizeof buf, f ) ) break;
	rc = sscanf( buf, " [%04X.%a[^]]] ", &serv_num, &serv_desc );
	if ( rc == 2 ) {
		desc.serv_num = serv_num;
		desc.serv_desc = serv_desc;
		serv_list.push_back( desc );
		continue;
	}
// strip white space from start and end of string
	p = buf;
	while ( *p == ' ' || *p == '\r' || *p == '\n' ) p++;
	if ( p != buf ) memmove( buf, p, strlen(p) + 1 );
	p = buf + strlen(buf) - 1;
	while ( p > buf && (*p == ' ' || *p == '\r' || *p =='\n' )) p--;
	p[1] = 0;

	if ( *buf != ';' && strlen(buf) && serv_list.size() )
		serv_list.back().func.push_back( strdup(buf) );
  }
  fclose( f );
}

csh capstone_ctx;
bool capstone_ctx_opened = false;

void disasm_mode_init() {
  int rc;

  if ( capstone_ctx_opened ) {
	cs_close( &capstone_ctx );
	capstone_ctx_opened = false;
  }

  if ( x86_mode )
	rc = cs_open( CS_ARCH_X86, CS_MODE_32, &capstone_ctx );
  else if ( x86_64_mode )
	rc = cs_open( CS_ARCH_X86, CS_MODE_64, &capstone_ctx );
  else if ( arm_mode )
	rc = cs_open( CS_ARCH_ARM, CS_MODE_ARM, &capstone_ctx );
  else return;

  if ( rc != CS_ERR_OK ) return;

  capstone_ctx_opened = true;

  cs_option( capstone_ctx, CS_OPT_SKIPDATA, CS_OPT_ON );
  cs_option( capstone_ctx, CS_OPT_DETAIL,   CS_OPT_ON );
}

int disassemble_capstone( int offs, char *buf, int buf_len, struct instr *insn ) {
  cs_insn *instr;

  size_t count = cs_disasm( capstone_ctx, (uint8_t*)buf, buf_len > 20? 20: buf_len, offs, 0, &instr );
  if ( !count ) return 0;

  sprintf( insn->mnemonic, "%-8s %s", instr[0].mnemonic, instr[0].op_str );

  int instr_size = instr[0].size;
  cs_free( instr, count );

  return instr_size;
}

bool relop_callback_called;
int insn_offset;

void relative_op_callback( char *buf, int blen, int op ) {
  relop_callback_called = true;
  snprintf( buf, blen, "0x%08x", insn_offset + op );
}

extern int disasm_c166( int offs, unsigned char *code, int len, char *instr_text );

int disassemble_address_wrapper_c166( int offs, char *buf, int buf_len, struct instr *insn ) {
  if ( offs & 1 ) return 0;
  return disasm_c166( offs, (unsigned char*)buf, buf_len, insn->mnemonic );
}

extern int or1k_disasm( unsigned int offs, unsigned char *code, int len, const char **mnemonic, const char **operands );

int disassemble_address_wrapper_or1k( int offs, char *buf, int buf_len, struct instr *insn ) {
  const char *mn, *op;

  if ( offs & 3 ) return 0;

  int rc = or1k_disasm( offs, (unsigned char*)buf, buf_len, &mn, &op );
  if ( rc > 0 ) sprintf( insn->mnemonic, "%-8s %s", mn, op );

  return rc;
}

int disassemble_address_wrapper( int offs, char *buf, int buf_len, struct instr *insn ) {
  if ( !buf || !buf_len ) return 0;
  if ( x86_mode || x86_64_mode || arm_mode )
	return disassemble_capstone( offs, buf, buf_len, insn );
  else if ( c166_mode )
  	return disassemble_address_wrapper_c166( offs, buf, buf_len, insn );
  else
	return disassemble_address_wrapper_or1k( offs,  buf, buf_len, insn );
}

struct le_obj_entry {
  int vseg_size, rel_base, flags;
  int pm_index, pm_num;
  char name[5];
};

union obj_entry {
  le_obj_entry le;
  pe_obj_entry pe;
};

vector <obj_entry> obj_table;
int le_hdr_offs;

bool load_le_objtbl() {
  unsigned char *p;
  int hdr_offs, obj_offs, objnum, i, *pd;
  obj_entry ent;

  le_hdr_offs = -1;

  p = get_data( 0x3c );
  if ( !p ) return false;
  hdr_offs = *(int*)p;
  if ( hdr_offs < 0 || hdr_offs >= fsize ) return false;
  p = get_data( hdr_offs );
  if ( memcmp( p, "LE", 2 ) ) return false;
  p = get_data( hdr_offs + 0x40 );
  if ( !p ) return false;
  
  obj_offs = *(int*)p;
  objnum = *(int*)(p+4);
  for ( i = 0; i < objnum; i++ ) {
  	pd = (int*) get_data( hdr_offs + obj_offs + i * 0x18 );
	ent.le.vseg_size = pd[0];
	ent.le.rel_base = pd[1];
	ent.le.flags = pd[2];
	ent.le.pm_index = pd[3];
	ent.le.pm_num = pd[4];
	memcpy( ent.le.name, &pd[5], 4 );
	ent.le.name[4] = 0;
	obj_table.push_back( ent );
  }

  le_hdr_offs = hdr_offs;
  return true;
}

bool load_pe_objtbl() {
  int i, j, hdr_offs, objtbl_offs, obj_offs;
  unsigned char *p;
  pe_header hdr;
  obj_entry ent;

  for ( i = 0; i < 40; i++ ) {
  	if ( (i << 3) + 4 >= fsize ) return false;
  	p = get_data( i << 3 );
        if ( !memcmp( p, "PE\0\0", 4 ) ) break;
  }
  if ( i == 40 ) return false;

  hdr_offs = i << 3;
  for ( i = 0; i < sizeof hdr; i++ ) {
  	p = get_data( hdr_offs + i );
	if ( !p ) return false;
  	((unsigned char*)&hdr)[i] = *p;
  }

  objtbl_offs = hdr_offs + hdr.nthdr_size + 0x18;
  
  for ( i = 0; i < hdr.num_objects; i++ ) {
  	obj_offs = objtbl_offs + i * sizeof(pe_obj_entry);
  	for ( j = 0; j < sizeof(pe_obj_entry); j++ ) {
	  	p = get_data( obj_offs + j );
		if ( !p ) return false;
	  	((unsigned char*)&ent)[j] = *p;
	}
	ent.pe.rva += hdr.image_base;
	obj_table.push_back( ent );
  }

  return true;
}
  
bool elf_32bits = false;
int elf_arch = EM_386;

bool load_elf_objtbl() {
  Elf *elf;
  Elf32_Ehdr *hdr32 = 0;
  Elf64_Ehdr *hdr64 = 0;
  int kind;

  elf_version( EV_CURRENT );
  elf = elf_begin( fd, ELF_C_READ, 0 );
  kind = elf_kind( elf );

  hdr32 = elf32_getehdr( elf );
  if ( !hdr32 ) hdr64 = elf64_getehdr( elf );

  elf_end( elf );

  if ( hdr32 ) elf_arch = hdr32->e_machine;
  else if ( hdr64 ) elf_arch = hdr64->e_machine;

  elf_32bits = hdr32;
  return kind == ELF_K_ELF;
}

struct select_list_ent {
  int x, y;
  char *str;
  void *data;
};

struct sel_list_state {
  WINDOW *w;
  vector <select_list_ent> lst;
  int attr1, attr2, lst_offs, ce, maxx, maxy;
};

inline void show_sel_list_entry( sel_list_state *s, int attr, int ent ) {
  wattrset( s->w, attr );
  mvwaddstr( s->w, s->lst[ent].y - s->lst_offs, s->lst[ent].x, s->lst[ent].str );
}

void display_sel_list( sel_list_state *s ) {
  int lst_offs = s->lst_offs;
  vector <select_list_ent> &lst = s->lst;
  int vsize = s->maxy - lst[0].y;

  for ( int i = 0; i < vsize && lst_offs + i < lst.size(); i++ ) {
  	if ( lst_offs + i == s->ce ) wattrset( s->w, s->attr1 );
	else wattrset( s->w, s->attr2 );
	mvwaddstr( s->w, lst[lst_offs + i].y - lst_offs,
		lst[lst_offs + i].x, lst[lst_offs + i].str );
  }

  if ( lst.size() > vsize ) {
  	int start, end;
  	start = lst_offs * vsize / lst.size();
	end = start + vsize * vsize / lst.size();
  	for ( int i = 0; i < s->maxy - lst[0].y; i++ ) {
		if ( i < start || i > end ) wattrset( s->w, s->attr2 );
		else wattrset( s->w, s->attr1 );
		mvwaddch( s->w, lst[lst_offs + i].y - lst_offs,
			s->maxx, ACS_CKBOARD );
	}
  }
}

int do_list_selection( WINDOW *w, int attr1, int attr2,
	vector <select_list_ent> &lst, void (*key_handler)(int key, void *data) = 0 ) {
  int i, c;
  sel_list_state s;

  if ( !lst.size() ) return -1;

  s.w = w;
  s.attr1 = attr1;
  s.attr2 = attr2;
  s.lst = lst;
  getmaxyx( w, s.maxy, s.maxx );
  s.maxy--;
  s.maxx--;
  s.ce = 0;
  s.lst_offs = 0;
  display_sel_list( &s );
  for (;;) {
  	wrefresh( w );
	move( 0, COLS-1 );
	refresh();
  	c = wgetch( w );
	if ( c == KEY_UP ) {
		if ( s.ce != 0 ) {
			show_sel_list_entry( &s, attr2, s.ce );
			s.ce--;
			if ( s.ce < s.lst_offs ) {
				s.lst_offs--;
				display_sel_list( &s );
			}
			else show_sel_list_entry( &s, attr1, s.ce );
		}
	}
	else if ( c == KEY_DOWN ) {
		if ( s.ce < lst.size()-1 ) {
			show_sel_list_entry( &s, attr2, s.ce );
			s.ce++;
			if ( s.lst[s.ce].y - s.lst_offs > s.maxy - 1 ) {
				s.lst_offs++;
				display_sel_list( &s );
			}
			else show_sel_list_entry( &s, attr1, s.ce );
		}
	}
	else if ( c == KEY_PPAGE ) {
		if ( s.ce != s.lst_offs ) {
			show_sel_list_entry( &s, attr2, s.ce );
			s.ce = s.lst_offs;
			show_sel_list_entry( &s, attr1, s.ce );
		}
		else {
			s.ce -= s.maxy - s.lst[0].y;
			if ( s.ce < 0 ) s.ce = 0;
			s.lst_offs = s.ce;
			display_sel_list( &s );
		}
	}
	else if ( c == KEY_NPAGE ) {
		int last_ent = s.lst_offs + s.maxy - s.lst[0].y - 1;
		if ( last_ent > s.lst.size() - 1 ) last_ent = s.lst.size()-1;
		if ( s.ce != last_ent ) {
			show_sel_list_entry( &s, attr2, s.ce );
			s.ce = last_ent;
			show_sel_list_entry( &s, attr1, s.ce );
		}
		else {
			s.ce += s.maxy - s.lst[0].y;
			if ( s.ce > s.lst.size() - 1 ) s.ce = s.lst.size()-1;
			s.lst_offs = s.ce - (s.maxy - s.lst[0].y - 1);
			if ( s.lst_offs < 0 ) s.lst_offs = 0;
			display_sel_list( &s );
		}
	}
	else if ( c == KEY_HOME ) {
		s.ce = 0;
		s.lst_offs = 0;
		display_sel_list( &s );
	}
	else if ( c == KEY_END ) {
		if ( lst.size() ) {
			s.ce = lst.size()-1;
			s.lst_offs = s.ce - (s.maxy - s.lst[0].y - 1);
			if ( s.lst_offs < 0 ) s.lst_offs = 0;
			display_sel_list( &s );
		}
	}
	else if ( c == '\n' ) return s.ce;
	else if ( c == 033 ) return -1;
	else if ( c == ('L' & 0x1f) ) {
		touchwin( stdscr );
		clearok( stdscr, true );
		refresh();
		touchwin( w );
	}
	else if ( key_handler )
		key_handler( c, lst[s.ce].data );
  }
}

int do_wlist_selection( int width, int height, const char *title, const char *tbl_header,
	vector<select_list_ent>& lst ) {

  WINDOW *w = newwin( height, width, (LINES-height)/2, (COLS-width)/2 );

  push_keyhints();
  clear_keyhints();

  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, title );
  mvwaddstr( w, 1, 1, tbl_header );

  int rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );

  delwin( w );
  pop_winstack();
  pop_keyhints();

  return rc;
}

int do_wlist_selection( int width, int height, const char *title, const char *tbl_header,
	vector<select_list_ent>& lst, void (*key_handler)(int key, void *data) ) {

  WINDOW *w = newwin( height, width, (LINES-height)/2, (COLS-width)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, title );
  mvwaddstr( w, 1, 1, tbl_header );

  int rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst, key_handler );

  delwin( w );
  pop_winstack();

  return rc;
}

int (*view_objtbl)() = 0;

int view_le_objtbl() {
  WINDOW *w;
  int i, c, rc, offs;
  unsigned char *p;
  select_list_ent sle;
  vector <select_list_ent> lst;

  push_keyhints();
  clear_keyhints();

  w = newwin( 15, 51, (LINES-15)/2, (COLS-51)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Object Table" );
  mvwaddstr( w, 1, 1, "Name RelBase  SegSize  PMIndex  PMNumber Flags" );
  for ( i = 0; i < obj_table.size(); i++ ) {
  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%4.4s %08x %08x %08x %08x %08x",
		obj_table[i].le.name, obj_table[i].le.rel_base,
		obj_table[i].le.vseg_size, obj_table[i].le.pm_index,
		obj_table[i].le.pm_num, obj_table[i].le.flags );
	lst.push_back( sle );
  }
  rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );
  if ( rc != -1 && lst.size() ) {
  	int dataoffs, pgidx, page, pgsize;

  	p = get_data( le_hdr_offs + 0x80 );
	dataoffs = *(int*)p;
	p = get_data( le_hdr_offs + 0x28 );
	pgsize = *(int*)p;
	p = get_data( le_hdr_offs + 0x48 );
	pgidx = le_hdr_offs + *(int*)p + (obj_table[rc].le.pm_index - 1) * 4;
	p = get_data( pgidx );
	page = *(unsigned short*)p + *(unsigned short*)(p + 2) - 1;
	offs = dataoffs + page * pgsize;
  }
  else offs = -1;

  delwin( w );
  pop_winstack();
  pop_keyhints();

  return offs;
}

int view_pe_objtbl() {
  WINDOW *w;
  int i, c, rc, offs;
  unsigned char *p;
  select_list_ent sle;
  vector <select_list_ent> lst;

  push_keyhints();
  clear_keyhints();

  w = newwin( 15, 55, (LINES-15)/2, (COLS-55)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Object Table" );
  mvwaddstr( w, 1, 1, "Name     VirtSize RVA      PhysSize PhysOffs Flags   " );
  for ( i = 0; i < obj_table.size(); i++ ) {
  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%-8.8s %08x %08x %08x %08x %08x",
		obj_table[i].pe.name, obj_table[i].pe.virtual_size,
		obj_table[i].pe.rva, obj_table[i].pe.physical_size,
		obj_table[i].pe.physical_offset, obj_table[i].pe.object_flags );
	lst.push_back( sle );
  }
  rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );

  if ( rc != -1 && lst.size() )
	offs = obj_table[rc].pe.physical_offset;
  else offs = -1;

  delwin( w );
  pop_winstack();
  pop_keyhints();

  return offs;
}

char *do_enter_filename() {
  WINDOW *w;
  static char buf[40];
  string_ent_state st;
  int c;

  w = newwin( 3, 40, (LINES-3)/2, (COLS-40)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(4) );
  mvwprintw( w, 0, 2, "Filename" );
  
  wmove( w, 1, 1 );
  init_cstr_entry( &st, w, (unsigned char*)buf, 38 );
  for (;;) {
	enter_cstring( &st, c, false );
	if ( c == 033 || c == '\n' ) break;
  }
  delwin( w );
  pop_winstack();
  if ( c == 033 ) return 0;

  return buf;
}

void objtbl_save_section( Elf_Scn *scn ) {
  char *fname = do_enter_filename();
  if ( !fname ) return;

  Elf_Data *d = elf_getdata( scn, 0 );
  FILE *f = fopen( fname, "w" );
  if ( f ) {
	fwrite( d->d_buf, 1, d->d_size, f );
	fclose( f );
  }
}

void objtbl_keyhandler( int c, void *data );
void elfsect_keyhandler( int c, void *data );

namespace view_elf32 {
#include "view_elf32.cc"
}

namespace view_elf64 {
#include "view_elf64.cc"
}

void objtbl_view_dyntab() {
  if ( elf_32bits )
	view_elf32::objtbl_view_dyntab();
  else
	view_elf64::objtbl_view_dyntab();
}

void objtbl_view_verobj() {
  if ( elf_32bits )
	view_elf32::objtbl_view_verobj();
  else
	view_elf64::objtbl_view_verobj();
}

void objtbl_view_verdobj() {
  if ( elf_32bits )
	view_elf32::objtbl_view_verdobj();
  else
	view_elf64::objtbl_view_verdobj();
}

int view_elf_objtbl() {
  if ( elf_32bits )
	return view_elf32::view_elf_objtbl();
  else
	return view_elf64::view_elf_objtbl();
}

void objtbl_keyhandler( int c, void *data ) {
  if ( c == KEY_F(2) ) objtbl_save_section( (Elf_Scn*)data );
  else if ( c == KEY_F(6) ) objtbl_view_dyntab();
  else if ( c == KEY_F(7) ) objtbl_view_verobj();
  else if ( c == KEY_F(8) ) objtbl_view_verdobj();
}

void elfsect_save_section( Elf32_Phdr *phdr ) {
  char *fname = do_enter_filename();
  if ( !fname ) return;

  int sz = phdr->p_filesz;
  char *buf = (char*) malloc( sz );

  lseek( fd, phdr->p_offset, SEEK_SET );
  sz = read( fd, buf, sz );

  FILE *f = fopen( fname, "w" );
  if ( f ) {
	fwrite( buf, 1, sz, f );
	fclose( f );
  }

  free( buf );
}

const char *elf_note_type( int type ) {
  switch ( type ) {
  case NT_PRSTATUS: return "NT_PRSTATUS";
  case NT_PRPSINFO: return "NT_PRPSINFO";
  case NT_SIGINFO:  return "NT_SIGINFO";
  case NT_AUXV:     return "NT_AUXV";
  case NT_FILE:     return "NT_FILE";
  }

  static char hex[12];
  sprintf( hex, "%08x", type );

  return hex;
}

unsigned int te32( unsigned int v ) {
  return (v >> 24) | ((v >> 8) & 0xff00) | ((v << 8) & 0xff0000) | (v << 24);
}

struct DisElf32_Nhdr {
  uint32_t n_type;
  string name;
  uint32_t data_offs;
  uint32_t data_len;
};

void elfprstatus_save( DisElf32_Nhdr *nhdr ) {
  char *fname = do_enter_filename();
  if ( !fname ) return;

  int sz = nhdr->data_len;
  char *buf = (char*) malloc( sz );

  lseek( fd, nhdr->data_offs, SEEK_SET );
  sz = read( fd, buf, sz );

  FILE *f = fopen( fname, "w" );
  if ( f ) {
	fwrite( buf, 1, sz, f );
	fclose( f );
  }

  free( buf );
}

void view_elf_prstatus( DisElf32_Nhdr *nhdr ) {
  vector <select_list_ent> lst;
  select_list_ent sle;
  int i;

  elf_prstatus *pr = (elf_prstatus*)get_data(nhdr->data_offs);
  for ( i = 0; i < 32; i++ ) {
  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "  r%-2d %08x  ", i, te32(pr->pr_reg.gpr[i]) );
	sle.data = (void*)pr->pr_reg.gpr[i];
	lst.push_back( sle );
  }

  sle.x = 1;
  sle.y = i + 2;
  asprintf( &sle.str, "  pc  %08x  ", te32(pr->pr_reg.pc) );
  sle.data = (void*)pr->pr_reg.pc;
  lst.push_back( sle );
  i++;

  sle.x = 1;
  sle.y = i + 2;
  asprintf( &sle.str, "  sr  %08x  ", te32(pr->pr_reg.sr) );
  sle.data = (void*)pr->pr_reg.sr;
  lst.push_back( sle );

  do_wlist_selection( 18, 20, "Registers",
  	"  Reg Value", lst );
}

void elfnote_keyhandler( int c, void *data ) {
  DisElf32_Nhdr *nhdr = (DisElf32_Nhdr*)data;
  if ( c == KEY_F(2) ) elfprstatus_save( nhdr );
  else if ( c == KEY_F(3) ) {
  	if ( nhdr->n_type == NT_PRSTATUS ) view_elf_prstatus( nhdr );
  }
}

void view_elf_note( Elf32_Phdr *phdr ) {
  vector <select_list_ent> lst;
  select_list_ent sle;
  vector <DisElf32_Nhdr> nhdr_lst;
  DisElf32_Nhdr nhdr;

  nhdr_lst.reserve( phdr->p_filesz / sizeof(Elf32_Nhdr) );

  int offs = 0, i = 0;
  while ( offs < phdr->p_filesz ) {
	Elf32_Nhdr *nh = (Elf32_Nhdr*)get_data( phdr->p_offset + offs );
        uint32_t n_type = te32(nh->n_type),
        	 n_namesz = te32(nh->n_namesz),
                 n_descsz = te32(nh->n_descsz);

	string name = string( ((char*)nh) + sizeof(Elf32_Nhdr), n_namesz );
	n_namesz = (n_namesz + 3) & ~3;

	nhdr.n_type = n_type;
	nhdr.name = name;
	nhdr.data_offs = phdr->p_offset + offs + sizeof(Elf32_Nhdr) + n_namesz;
	nhdr.data_len = n_descsz;
	nhdr_lst.push_back( nhdr );

  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%-11.11s %-10.10s %08x %08x",
		elf_note_type( n_type ), name.c_str(),
		nhdr.data_offs - phdr->p_offset, n_descsz );
	sle.data = &nhdr_lst.back();
	lst.push_back( sle );

	n_descsz = (n_descsz + 3) & ~3;
	offs += sizeof(Elf32_Nhdr) + n_namesz + n_descsz;

	i++;
  }

  push_keyhints();
  clear_keyhints();
  set_keyhint( 2, "SavDat" );
  set_keyhint( 3, "Detail" );

  do_wlist_selection( 42, 15, "Elf Note Table",
  	"Type        Name       DataOffs DataSz",
	lst, elfnote_keyhandler );

  pop_keyhints();
}

void elfsect_keyhandler( int c, void *data ) {
  Elf32_Phdr *phdr = (Elf32_Phdr*)data;
  if ( c == KEY_F(2) ) elfsect_save_section( phdr );
  else if ( c == KEY_F(3) ) {
  	if ( elf_arch == EM_OPENRISC && phdr->p_type == PT_NOTE )
		view_elf_note( phdr );
  }
}

int do_search( int offs0, unsigned char *pat, int patlen, bool forw ) {
  int i, offs = offs0;
  unsigned char *buf, *p;

  buf = (unsigned char*) malloc( patlen );
  for ( i = 0; i < patlen; i++ ) {
  	p = get_data( offs + i );
	if ( !p ) {
		free( buf );
		return -1;
	}
	buf[i] = *p;
  }

  for (;;) {
  	if ( !memcmp( pat, buf, patlen ) ) break;

	if ( forw ) {
		memmove( buf, buf+1, patlen-1 );
		p = get_data( offs + patlen );
	}
	else {
		memmove( buf+1, buf, patlen-1 );
		p = get_data( offs-1 );
	}

	if ( !p ) {
		free( buf );
		return -1;
	}
	
	if ( forw ) {
		buf[patlen-1] = *p;
		offs++;
	}
	else {
		*buf = *p;
		offs--;
	}
  }

  free( buf );
  return offs;
}

void do_showabout() {
  WINDOW *win;
  int c;

  push_keyhints();
  clear_keyhints();
  win = newwin( 6, 51, (LINES-6)/2, (COLS-51)/2 );
  box( win, ACS_VLINE, ACS_HLINE );
  wbkgd( win, COLOR_PAIR(4) );
  mvwprintw( win, 1, 1, "Disassembler v0.3" );
  mvwprintw( win, 2, 1, "(c) 2003-2018 Copyright Vad Rulezz <vr5@narod.ru>" );
  mvwprintw( win, 3, 1, "Homepage: http://vr5.narod.ru/disasm" );
  mvwprintw( win, 4, 1, "Licensed under GPL v2" );
  wrefresh( win );
  move( 0, COLS-1 );
  refresh();
  for (;;) {
	c = wgetch( win );
	if ( c != ('L' & 0x1f) ) break;
	clearok( stdscr, true );
	refresh();
  }
  delwin( win );
  pop_keyhints();
}

static int hex_row_bytes;

void init_hex() {
  hex_row_bytes = (COLS - 9) / (8 * 3 + 2 + 8) * 8; // 8 hexbytes + 8 ascii-bytes + decorations
}

void do_showhex( int offs ) {
  int i, j, k;
  unsigned char *p;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = 0; i < LINES-2; i++ ) {
  	mvprintw( i+1, 0, "%08X  ", offs );
	for ( j = 0; j < 8; j++ ) {
		p = get_data( offs + j );
		if ( p ) printw( "%02X ", *p );
		else addstr( "   " );
	}
	for ( k = 1; k < hex_row_bytes / 8; k++ ) {
		addstr( "- " );
		for ( j = 0; j < 8; j++ ) {
			p = get_data( offs + k * 8 + j );
			if ( p ) printw( "%02X ", *p );
			else addstr( "   " );
		}
	}
	addch( ' ' );
	for ( j = 0; j < hex_row_bytes; j++ ) {
		p = get_data( offs + j );
		if ( !p ) {
			addch( ' ' );
			continue;
		}
		if ( isprint( *p ) && *p != 0x9b )
//		if ( *p >= 32 && *p < 128 )
			addch( *p );
		else
			addch( '.' );
	}
	offs += hex_row_bytes;
  }
}

void do_showtext( int offs ) {
  int i, j, rc;
  unsigned char *p;
#ifdef USE_WIDEC
  wchar_t wc[2];
  cchar_t cc;
  char mbs[MB_CUR_MAX+1];

  wc[1] = 0;
  setlocale( LC_CTYPE, mb_locale );
#endif
 
  attrset( COLOR_PAIR(2) | A_BOLD );
  move( 1, 0 );
  for ( i = 0; i < LINES-2; i++ )
  	for ( j = 0; j < COLS; j++ ) {
		p = get_data( offs );
		if ( !p ) {
			addch( ' ' );
			offs++;
			continue;
		}
#ifdef USE_WIDEC
		rc = mbtowc( &wc[0], (char*)p, 1 );
		if ( rc != -1 && iswprint( wc[0] ) )
			addwstr( wc );
#else
		if ( *p != 0x9b && isprint( *p ) )
			addch( *p );
#endif
		else addch( '.' );
		offs++;
	}
#ifdef USE_WIDEC
  setlocale( LC_CTYPE, "" );
#endif
  refresh();
}

void do_highlight_text( int offs, int patoffs, int patlen ) {
  int i;
  unsigned char *p;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ||
       patoffs > offs + (LINES-2)*COLS ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  if ( patoffs + patlen >= offs + (LINES-2)*COLS )
  	patlen = offs + (LINES-2)*COLS - patoffs;
  attrset( COLOR_PAIR(3) | A_REVERSE );
  move( (patoffs-offs)/COLS+1, (patoffs-offs)%COLS );
  for ( i = 0; i < patlen; i++ ) {
  	p = get_data( patoffs + i );
	if ( !p ) {
		addch( ' ' );
		continue;
	}
  	if ( *p != 0x9b && isprint(*p) ) addch( *p ); else addch( '.' );
  }
  move( 0, COLS-1 );
}
		
void do_highlight_hex( int offs, int patoffs, int patlen ) {
  unsigned char *p;
  int i, line, pos, poffs;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ||
       patoffs > offs + (LINES-2) * hex_row_bytes ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  if ( patoffs + patlen >= offs + (LINES-2) * hex_row_bytes )
  	patlen = offs + (LINES-2) * hex_row_bytes - patoffs;
  attrset( COLOR_PAIR(3) | A_REVERSE );
  for ( i = 0; i < patlen; i++ ) {
  	line = (patoffs + i - offs) / hex_row_bytes + 1;
	pos  = (patoffs + i - offs) % hex_row_bytes;
  	p = get_data( patoffs + i );
	poffs = 10 + pos / 8 * 2 + pos * 3;
	mvprintw( line, poffs, "%02X", *p );
	if ( i != patlen - 1 ) {
		if ( pos != hex_row_bytes - 1 ) {
			addch( ' ' );
			if ( pos % 8 == 7 ) addstr( "- " );
		}
	}

	poffs = 10 + hex_row_bytes * 3 + (hex_row_bytes / 8 - 1) * 2 + 1;
	if ( *p != 0x9b && isprint( *p ) ) mvaddch( line, pos + poffs, *p );
	else mvaddch( line, pos + poffs, '.' );
  }

  move( 0, COLS-1 );
}

void do_highlight_disasm( int offs, int patoffs, int patlen ) {
  int sz, i, j;
  struct instr insn;
  unsigned char *p;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  attrset( COLOR_PAIR(3) | A_REVERSE );
  for ( i = 0; i < LINES-2; i++ ) {
  	if ( offs > patoffs + patlen ) break;
	p = get_data( offs );
	sz = disassemble_address_wrapper( offs, (char*)p, fsize - offs, &insn );
	if ( !sz ) sz = 1;
	for ( j = 0; j < sz; j++ )
		if ( offs + j >= patoffs && offs + j < patoffs + patlen )
			mvprintw( i + 1, j * 2 + 9, "%02X", p[j] );
	offs += sz;
  }
  move( 0, COLS-1 );
}

bool quit_requested;

void edit_hexcode( int offs ) {
  bool hi_nibble, ascii_mode;
  char istr[2];
  unsigned char *buf, *p;
  int c, val, val2, coffs, x, y, i, xtbl[hex_row_bytes];
  int asc_pos;

  x = 10;
  for ( i = 0; i < hex_row_bytes; i++ ) {
  	xtbl[i] = x;
	x += 3;
	if ( i % 8 == 7 ) x += 2;
  }
  asc_pos = 10 + hex_row_bytes * 3 + (hex_row_bytes / 8 - 1) * 2 + 1;

  offs &= ~0x0f;

  buf = (unsigned char*) malloc( hex_row_bytes * (LINES-2) );
  for ( i = 0; i < hex_row_bytes * (LINES-2); i++ ) {
  	p = get_data( offs + i );
	if ( p ) buf[i] = *p; else buf[i] = 0;
  }

  push_keyhints();
  clear_keyhints();
  set_keyhint( 9, "Save" );
  set_keyhint( 10, "Quit" );

  hi_nibble = true;
  ascii_mode = false;
  coffs = 0;
  for (;;) {
  	y = coffs/hex_row_bytes + 1;
	if ( ascii_mode )
		x = asc_pos + coffs % hex_row_bytes;
	else {
		x = xtbl[coffs % hex_row_bytes];
		if ( !hi_nibble ) x++;
	}
	move( y, x );

get_next_key:
	c = getch();
	if ( c == KEY_RIGHT ) {
move_to_right:
		if ( ascii_mode ) {
			if ( coffs != (LINES-2) * hex_row_bytes - 1 ) coffs++;
		}
		else {
			if ( !hi_nibble ) {
				if ( coffs == (LINES-2) * hex_row_bytes - 1 )
					goto get_next_key;
				coffs++;
			}
			hi_nibble = !hi_nibble;
		}	
	}
	else if ( c == KEY_LEFT ) {
		if ( ascii_mode ) {
			if ( coffs > 0 ) coffs--;
		}
		else {
			if ( hi_nibble ) {
				if ( !coffs ) goto get_next_key;
				coffs--;
			}
			hi_nibble = !hi_nibble;
		}
	}
	else if ( c == KEY_DOWN ) {
		if ( coffs + hex_row_bytes >= (LINES-2) * hex_row_bytes ) goto get_next_key;
		coffs += hex_row_bytes;
	}
	else if ( c == KEY_UP ) {
		if ( coffs < hex_row_bytes ) goto get_next_key;
		coffs -= hex_row_bytes;
	}
	else if ( c == '\t' ) {
		ascii_mode = !ascii_mode;
		if ( !ascii_mode ) hi_nibble = true;
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( coffs == 0 ) goto get_next_key;

		if ( ascii_mode ) {
			coffs--;
			val = *get_data( offs + coffs );
			buf[coffs] = val;
			attrset( COLOR_PAIR(2) | A_BOLD );
			mvprintw( coffs / hex_row_bytes + 1, xtbl[coffs % hex_row_bytes], "%02X", val );
			move( coffs / hex_row_bytes + 1, asc_pos + coffs % hex_row_bytes );
			if ( isprint(val) && val != 0x9b ) addch( val ); else addch( '.' );
		}
		else {
			if ( hi_nibble ) coffs--;
			hi_nibble = !hi_nibble;
			val = *get_data( offs + coffs );
			val2 = buf[coffs];
			attrset( COLOR_PAIR(2) | A_BOLD );
			if ( hi_nibble ) {
				buf[coffs] = val2 & 0xf | val & 0xf0;
				mvprintw( coffs / hex_row_bytes + 1, xtbl[coffs % hex_row_bytes], "%X", val >> 4 );
			}
			else {
				buf[coffs] = val2 & 0xf0 | val & 0xf;
				mvprintw( coffs / hex_row_bytes + 1, xtbl[coffs % hex_row_bytes] + 1, "%X", val & 0xf );
			}
			if ( buf[coffs] != val ) attrset( COLOR_PAIR(5) | A_BOLD );
			else attrset( COLOR_PAIR(2) | A_BOLD );
			move( coffs / hex_row_bytes + 1, asc_pos + coffs % hex_row_bytes );
			if ( isprint(buf[coffs]) && buf[coffs] != 0x9b )
				addch( buf[coffs] );
			else addch( '.' );
		}
	}
	else if ( c == KEY_F(9) ) {
		int fdwr, rc;

		fdwr = open( fname, O_RDWR );
		if ( fdwr == -1 ) continue;

		rc = lseek( fdwr, offs, SEEK_SET );
		if ( rc == -1 ) continue;

		int len = hex_row_bytes * (LINES-2);
		if ( len > fsize - offs ) len = fsize - offs;

		rc = write( fdwr, buf, len );
		if ( rc == -1 ) continue;

		close( fdwr );

		buf_offs = -512;	// refetch data to read buffer
		break;
	}
	else if ( c == KEY_F(10) ) {
		quit_requested = true;
		break;
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
	else if ( c == 033 ) break;
	else if ( ascii_mode && c == (c & 0xff) ) {
		val2 = *get_data( offs + coffs );

		buf[coffs] = c;
		if ( val2 != c ) attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );

		if ( isprint( c ) && c != 0x9b ) addch( c );
		else addch( '.' );

		y = coffs / hex_row_bytes + 1;
		x = xtbl[coffs % hex_row_bytes];
		move( y, x );
		if ( (val2 & 0xf0) != (c & 0xf0) )
			attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
		printw( "%X", c >> 4 );
		if ( (val2 & 0xf) != (c & 0xf) )
			attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
		printw( "%X", c & 0xf );

		goto move_to_right;
	}
	else if ( !ascii_mode && c == (c & 0xff) && isxdigit(c) ) {
		bool change = false;

		istr[0] = c;
		istr[1] = 0;
		val = strtol( istr, 0, 16 );
		val2 = *get_data( offs + coffs );
	
		p = &buf[coffs];
		if ( hi_nibble ) {
			val = val << 4 | (*p & 0xf);
			if ( (val & 0xf0) != (val2 & 0xf0) ) change = true;
		}
		else {
			val |= *p & 0xf0;
			if ( (val & 0xf) != (val2 & 0xf) ) change = true;
		}
		*p = val;

		if ( change ) attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
	
		addch( toupper(c) );

		if ( val == val2 ) attrset( COLOR_PAIR(2) | A_BOLD );
		else attrset( COLOR_PAIR(5) | A_BOLD );
		move( coffs / hex_row_bytes + 1, asc_pos + coffs % hex_row_bytes );
		if ( isprint(buf[coffs]) && buf[coffs] != 0x9b )
			addch( buf[coffs] );
		else addch( '.' );

		goto move_to_right;
	}
  }
  free( buf );

  pop_keyhints();
  move( 0, 0 );
}

struct edit_disasm_state {
  int offs, line, pos, *offstbl;
  int bufsz;
  unsigned char *buf;
  bool hi_nibble;
};

void update_offstbl( edit_disasm_state& st ) {
  int i, coffs, sz;
  char *p;
  struct instr insn;
  
  coffs = st.offstbl[st.line];
  for ( i = st.line; i < LINES - 2; i++ ) {
  	st.offstbl[i] = coffs;
	if ( coffs >= st.bufsz - 20 ) {
		st.bufsz += 512;
		st.buf = (unsigned char*) realloc( st.buf, st.bufsz );
	}
  	sz = disassemble_address_wrapper( st.offs + coffs, (char*)&st.buf[coffs], st.bufsz - coffs, &insn );
	if ( !sz ) sz++;
	coffs += sz;
  }
  st.offstbl[i] = coffs;
}

void print_hexbyte_upd( int val, int val2 ) {
  if ( (val & 0xf0) != (val2 & 0xf0) ) {
	attrset( COLOR_PAIR(5) | A_BOLD );
	printw( "%X", val2 >> 4 );
	attrset( COLOR_PAIR(2) | A_BOLD );
  }
  else printw( "%X", val2 >> 4 );
  if ( (val & 0xf) != (val2 & 0xf) ) {
	attrset( COLOR_PAIR(5) | A_BOLD );
	printw( "%X", val2 & 0xf );
	attrset( COLOR_PAIR(2) | A_BOLD );
  }
  else printw( "%X", val2 & 0xf );
}

void update_bview( edit_disasm_state& st ) {
  int i, j, offs, val, val2, y, x, sz;
  unsigned char *p, *p2;
  struct instr insn;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = st.line; i < LINES-2; i++ ) {
  	memset( &insn, 0, sizeof insn );
	offs = st.offstbl[i];
	p = &st.buf[offs];
  	sz = disassemble_address_wrapper( st.offs + offs, (char*)p, st.bufsz - offs, &insn );
	mvprintw( i + 1, 0, "%08X ", st.offs + offs );
	if ( sz ) {
		for ( j = 0; j < 8; j++ )
			if ( j < sz ) {
				val = *get_data( st.offs + offs + j );
				val2 = st.buf[offs + j];
				print_hexbyte_upd( val, val2 );
			}
			else printw( "  " );
		printw( " %s", insn.mnemonic );
	}
	else {
		val = *get_data( st.offs + offs );
		val2 = st.buf[offs];
		print_hexbyte_upd( val, val2 );
		addstr( "               (bad)" );
	}
	getyx( stdscr, y, x );
	for ( j = x; j < COLS; j++ ) addch( ' ' );
  }
}

int do_translate_asm( int current_ip, unsigned char *codebuf, int cblen );

void edit_disasm( int offs ) {
  int c, i, val, val2, sz, coffs;
  unsigned char *p;
  char istr[2];
  bool change;
  struct instr insn;
  edit_disasm_state st;

  st.offstbl = (int*) malloc( (LINES - 1) * sizeof(int) );
  st.offs = offs;
  coffs = offs;
  for ( i = 0; i < LINES - 2; i++ ) {
  	st.offstbl[i] = coffs - offs;
  	p = get_data( coffs );
  	sz = disassemble_address_wrapper( coffs, (char*)p, fsize - coffs, &insn );
	if ( !sz ) sz++;
	coffs += sz;
  }
  st.offstbl[i] = coffs - offs;

  push_keyhints();
  clear_keyhints();
  set_keyhint( 7, "Asm" );
  set_keyhint( 9, "Save" );
  set_keyhint( 10, "Quit" );
 
  st.bufsz = (coffs - offs + 511) & ~511;
  st.buf = (unsigned char*) malloc( st.bufsz );
  for ( i = 0; i < st.bufsz; i++ ) {
  	p = get_data( offs + i );
	st.buf[i] = *p;
  }

  coffs = 0;
  st.hi_nibble = true;
  st.line = 0;
  st.pos = 0;
  for (;;) {
  	move( st.line+1, st.pos+9 );
get_next_key:
	c = getch();
	if ( c == KEY_DOWN ) {
move_down:
		if ( st.line+1 >= LINES-2 ) goto get_next_key;
		st.line++;
		if ( st.pos >= (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 ) {
			st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			st.hi_nibble = false;
		}
	}
	else if ( c == KEY_UP ) {
		if ( !st.line ) goto get_next_key;
		st.line--;
		if ( st.pos >= (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 ) {
			st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			st.hi_nibble = false;
		}
	}
	else if ( c == KEY_RIGHT ) {
move_to_right:
		if ( st.pos+1 < (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2) {
			st.pos++;
			st.hi_nibble = !st.hi_nibble;
		}
		else
			if ( st.line+1 < LINES-2 ) {
				st.line++;
				st.pos = 0;
				st.hi_nibble = true;
			}
	}
	else if ( c == KEY_LEFT ) {
		if ( st.pos ) {
			st.pos--;
			st.hi_nibble = !st.hi_nibble;
		}
		else
			if ( st.line ) {
				st.line--;
				st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
				st.hi_nibble = false;
			}
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( st.pos ) st.pos--;
		else {
			if ( st.line ) {
				st.line--;
				st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			}
			else goto get_next_key;
		}
		st.hi_nibble = !st.hi_nibble;
	
		coffs = st.offstbl[st.line] + st.pos/2;
		val = st.buf[coffs];
		val2 = *get_data( offs + coffs );
		if ( st.hi_nibble )
			st.buf[coffs] = val & 0xf | val2 & 0xf0;
		else
			st.buf[coffs] = val & 0xf0 | val2 & 0xf;

		update_offstbl( st );
		update_bview( st );
	}
	else if ( c == (c & 0xff) && isxdigit(c) ) {
		istr[0] = c;
		istr[1] = 0;
		val = strtol( istr, 0, 16 );

		coffs = st.offstbl[st.line] + st.pos/2;
		val2 = *get_data( offs + coffs );
		if ( st.hi_nibble )
			st.buf[coffs] = (st.buf[coffs] & 0xf) | (val << 4);
		else
			st.buf[coffs] = (st.buf[coffs] & 0xf0) | val;

		update_offstbl( st );
		update_bview( st );
		goto move_to_right;
	}
	else if ( c == KEY_F(7) ) {
		coffs = st.offstbl[st.line] + st.pos/2;
		do_translate_asm( offs + coffs,
			&st.buf[coffs], st.bufsz - coffs );
		update_offstbl( st );
		update_bview( st );
		goto move_down;
	}
	else if ( c == KEY_F(9) ) {
		int fdwr, rc;
		
		fdwr = open( fname, O_RDWR );
		if ( fdwr == -1 ) continue;
		
		rc = lseek( fdwr, offs, SEEK_SET );
		if ( rc == -1 ) continue;

		rc = write( fdwr, st.buf, st.bufsz );
		if ( rc == -1 ) continue;

		close( fdwr );

		buf_offs = -512;	// refetch data to read buffer
		int ln = st.line;
		st.line = 0;
		update_bview( st );
		st.line = ln;
		break;
	}
	else if ( c == 033 ) break;
	else if ( c == KEY_F(10) ) {
		quit_requested = true;
		break;
	}
  }

  pop_keyhints();

  free( st.buf );
  free( st.offstbl );
}

void do_disasm( int offs ) {
  int sz, i, j, x, y;
  unsigned char *p;
  struct instr insn;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = 0; i < LINES-2; i++ ) {
  	memset( &insn, 0, sizeof insn );
	p = get_data( offs );
	if ( !p ) break;
  	sz = disassemble_address_wrapper( offs, (char*)p, fsize - offs, &insn );
	if ( sz ) {
		mvprintw( i + 1, 0, "%08X ", offs );
		for ( j = 0; j < 8; j++ )
			if ( j < sz ) printw( "%02X", p[j] );
			else printw( "  " );
		printw( " %s", insn.mnemonic );
		offs += sz;
	}
	else {
		mvprintw( i + 1, 0, "%08X %02X               (bad)",
			offs, ((unsigned char*)p)[0] );
		offs++;
	}
	getyx( stdscr, y, x );
	for ( j = x; j < COLS; j++ ) addch( ' ' );
  }

  for ( ; i < LINES-2; i++ ) {
	move( i + 1, 0 );
	for ( j = 0; j < COLS; j++ ) addch( ' ' );
  }
}

void try_scroll_down( int& offs, int lines=1 ) {
  struct instr insn;
  unsigned char *p;
  int sz;

  while ( lines ) {
	p = get_data( offs );
	sz = disassemble_address_wrapper( offs, (char*)p, fsize - offs, &insn );
	if ( !sz && fsize - offs ) sz++;
	offs += sz;
	lines--;
  }
}

void try_scroll_up( int& offs, int lines=1 ) {
  int sz, i, best, offs2;
  struct instr insn;
  unsigned char *p;

  while ( offs && lines ) {
  	best = 1;

  	for ( i = 2; i < 9; i++ ) {
		offs2 = offs - i;
		if ( offs2 < 0 ) break;

		p = get_data( offs2 );
		sz = disassemble_address_wrapper( offs2, (char*)p, fsize - offs2, &insn );
		if ( sz == i ) best = i;
	}

	offs -= best;
	lines--;
  }
}

enum disasm_mode_type {
  DISASM_MODE,
  HEX_MODE,
  TEXT_MODE
};

disasm_mode_type disasm_mode;

void enter_offset( int& offs );

bool do_asm_str_x86( char *str );
extern unsigned char asmbuf_x86[50], *asmbuf_ptr_x86;
bool do_asm_str_arm( char *str );
extern unsigned char asmbuf_arm[50], *asmbuf_ptr_arm;
extern int current_ip;

bool do_asm_str( int current_ip0, char *str ) {
  if ( x86_mode )
  	return do_asm_str_x86( str );

  if ( arm_mode ) {
  	current_ip = current_ip0;
  	return do_asm_str_arm( str );
  }

  return false;
}

unsigned char *get_asm_data( int& datalen ) {
  if ( x86_mode ) {
  	datalen = asmbuf_ptr_x86 - asmbuf_x86;
	return asmbuf_x86;
  }
  else if ( arm_mode ) {
	datalen = asmbuf_ptr_arm - asmbuf_arm;
	return asmbuf_arm;
  }

  datalen = 0;
  return 0;
}

int do_translate_asm( int current_ip, unsigned char *codebuf, int cblen ) {
  WINDOW *w;
  unsigned char buf[40];
  string_ent_state st;
  int c;

  w = newwin( 3, 40, (LINES-3)/2, (COLS-40)/2 );
  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(4) );
  if ( x86_mode )
	mvwprintw( w, 0, 2, "x86 assembler" );
  else if ( arm_mode )
  	mvwprintw( w, 0, 2, "ARM assembler" );
  else if ( c166_mode )
  	mvwprintw( w, 0, 2, "c166 assembler" );
  else {
  	delwin( w );
  	return 0;
  }
  
  wmove( w, 1, 1 );
  init_cstr_entry( &st, w, buf, 38 );
  for (;;) {
	enter_cstring( &st, c, false );
	if ( c == 033 || c == '\n' ) break;
  }
  delwin( w );
  if ( c == 033 ) return -1;

  buf[st.idx] = 0;
  if ( !do_asm_str( current_ip, (char*)buf ) ) return -1;
  
  unsigned char *asmbuf;
  int asmlen;
  asmbuf = get_asm_data( asmlen );

  if ( asmlen < cblen ) cblen = asmlen;
  memcpy( codebuf, asmbuf, cblen );
  
  return cblen;
}

bool do_enter_search( int offs, unsigned char **spat,
	int *lpat, bool& forward ) {

  unsigned char *buf;
  WINDOW *win;
  int n, c, done, rc;
  string_ent_state xst, cst;
  bool ascii = false;

  push_keyhints();
  clear_keyhints();
  set_keyhint( 2, "Dir" );
  set_keyhint( 7, "Asm" );
  refresh();
 
  if ( !spat ) return false;
  if ( *spat )
  	buf = *spat;
  else
  	buf = *spat = (unsigned char*) malloc( 50 );
 
  win = newwin( 6, 50, (LINES-6)/2, (COLS-50)/2 );
  keypad( win, TRUE );
  box( win, ACS_VLINE, ACS_HLINE );
  wbkgd( win, COLOR_PAIR(4) );
  mvwprintw( win, 0, 2, "Search/" );
  waddstr( win, forward? "Forward": "Backward" );
  mvwprintw( win, 1, 1, "Ascii" );
  mvwprintw( win, 3, 1, "Hex" );
  wmove( win, 2, 1 );
  init_xstr_entry( &cst, win, buf, 16 );
  cst.idx = *lpat;
  update_cstring( &cst );
  wmove( win, 4, 1 );
  init_xstr_entry( &xst, win, buf, 16 );
  xst.idx = *lpat;
  update_xstring( &xst );
  for (;;) {
  	if ( ascii ) {
		done = enter_cstring( &cst, c, true );
		xst.idx = cst.idx;
		xst.hi_nibble = true;
		update_xstring( &xst );
		if ( done ) break;
	}
	else {
		done = enter_xstring( &xst, c, true );
		cst.idx = xst.idx;
		if ( !xst.hi_nibble ) cst.idx++;
		update_cstring( &cst );
		if ( done ) break;
	}

	if ( c == KEY_UP || c == KEY_DOWN || c == '\t' )
		ascii = !ascii;
	else if ( c == KEY_F(2) ) {
		forward = !forward;
		if ( forward ) {
			mvwaddstr( win, 0, 9, "Forward" );
			waddch( win, ACS_HLINE );
		}
		else
			mvwaddstr( win, 0, 9, "Backward" );
	}
	else if ( c == KEY_F(7) ) {
		rc = do_translate_asm( offs, buf + cst.idx, 16 - cst.idx );
		touchwin( win );
		wrefresh( win );
		if ( rc != -1 ) {
			cst.idx += rc;
			xst.idx += rc;
			update_cstring( &cst );
			update_xstring( &xst );
		}
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
  }
  delwin( win );

  pop_keyhints();

  *lpat = cst.idx;
  return done == 1;
}

void interactive_asm( int offs );

void disasm_loop( int& offs ) {
  int c, rc, patoffs = -1, lpat = 0;
  unsigned char *spat = 0;
  bool forw = true;

  push_keyhints();
  if ( !view_objtbl ) set_keyhint( 6, "      " );
//  set_keyhint( 2, "IntAsm" );
  set_keyhint( 2, "      " );

  if ( x86_mode )
	set_keyhint( 8, "x86_64" );
  else if ( x86_64_mode )
	set_keyhint( 8, "ArmIsa" );
  else if ( arm_mode )
	set_keyhint( 8, "166Isa" );
  else if ( c166_mode )
	set_keyhint( 8, "or1Isa" );
  else
	set_keyhint( 8, "x86Isa" );

  do_disasm( offs );

  while ( !quit_requested ) {
	c = getch();
	switch (c) {
	case KEY_DOWN:
		try_scroll_down( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_UP:
		try_scroll_up( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_NPAGE:
		try_scroll_down( offs, LINES-2 );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		try_scroll_up( offs, LINES-2 );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(2):
		interactive_asm( offs );
		do_disasm( offs );
		break;
	case KEY_F(3):
		edit_disasm( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = HEX_MODE;
		init_hex();
		goto leave_mode;
	case KEY_F(5):
		enter_offset( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(6):
		if ( view_objtbl ) {
			rc = view_objtbl();
			if ( rc != -1 ) offs = rc;
		}
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(8):
		if ( x86_mode ) {  // -> x86_64
			x86_mode = !x86_mode;
			x86_64_mode = !x86_64_mode;
  			set_keyhint( 8, "ArmIsa" );
		}
		else if ( x86_64_mode ) { // -> ARM
			x86_64_mode = !x86_64_mode;
			arm_mode = !arm_mode;
  			set_keyhint( 8, "166Isa" );
		}
		else if ( arm_mode ) { // -> C166
			arm_mode = !arm_mode;
			c166_mode = !c166_mode;
  			set_keyhint( 8, "or1Isa" );
		}
		else if ( c166_mode ) { // -> or1k
			c166_mode = !c166_mode;
			set_keyhint( 8, "x86Isa" );
		}
		else { //  -> x86
			x86_mode = !x86_mode;
  			set_keyhint( 8, "x86_64" );
		}
		disasm_mode_init();
		do_disasm( offs );
		break;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			int srch_from;
	
			srch_from = forw? patoffs+1: patoffs-1; 
			patoffs = do_search( srch_from, spat, lpat, forw );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( offs, &spat, &lpat, forw ) ) {
			patoffs = do_search( offs, spat, lpat, forw );
search_done:
			if ( patoffs != -1 ) offs = patoffs;
		}
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;;
	}
  }

leave_mode:
  pop_keyhints();
}

void hex_loop( int& offs ) {
  int c, rc, lpat = 0, patoffs = -1;
  unsigned char *spat = 0;
  bool forw = true;

  push_keyhints();
  if ( !view_objtbl ) set_keyhint( 6, "      " );

  offs &= ~(hex_row_bytes - 1);
  do_showhex( offs );

  while ( !quit_requested ) {
	c = getch();
	switch (c) {
	case KEY_DOWN:
		if ( offs + (LINES-2) * hex_row_bytes < fsize ) offs += hex_row_bytes;
//		else {
//			if ( fsize > (LINES-2) * 16 )
//				offs = fsize - (LINES-2) * 16;
//			else offs = 0;
//		}
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_UP:
		if ( offs >= hex_row_bytes ) {
			offs -= hex_row_bytes;
			do_showhex( offs );
			do_highlight_hex( offs, patoffs, lpat );
		}
		else offs = 0;
		break;
	case KEY_NPAGE:
		if ( offs + (LINES-2) * hex_row_bytes * 2 < fsize ) offs += (LINES-2) * hex_row_bytes;
		else {
			if ( fsize > (LINES-2) * hex_row_bytes ) {
				offs = (fsize - (LINES-2) * hex_row_bytes) & ~(hex_row_bytes - 1);
			}
			else offs = 0;
		}

		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		if ( offs > (LINES-2) * hex_row_bytes ) offs -= (LINES-2) * hex_row_bytes;
		else offs = 0;

		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_END:
		offs = (fsize - (LINES-2) * hex_row_bytes & ~(hex_row_bytes - 1));
		if ( offs < 0 ) offs = 0;
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(3):
		edit_hexcode( offs );
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = TEXT_MODE;
		goto leave_mode;
	case KEY_F(5):
		enter_offset( offs );
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(6):
		if ( view_objtbl ) {
			rc = view_objtbl();
			if ( rc != -1 ) offs = rc;
		}
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			int srch_from;
	
			srch_from = forw? patoffs+1: patoffs-1; 
			patoffs = do_search( srch_from, spat, lpat, forw );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( offs, &spat, &lpat, forw ) ) {
			patoffs = do_search( offs, spat, lpat, forw );
search_done:	
			if ( patoffs != -1 ) {
				if ( patoffs < offs ||
				     patoffs > offs + (LINES-2) * hex_row_bytes ) {
					offs = patoffs;
					offs -= offs % hex_row_bytes;
				}
			}
		}
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;
	}
  }
leave_mode:
  pop_keyhints();
}

void text_loop( int& offs ) {
  int c, patoffs = -1, lpat = 0;
  unsigned char *spat = 0;
  bool forw = true;

  push_keyhints();
  set_keyhint( 3, "      " );
  set_keyhint( 5, "      " );
  set_keyhint( 6, "      " );

  do_showtext( offs );

  while ( !quit_requested ) {
  	c = getch();
	switch ( c ) {
	case KEY_DOWN:
		offs += COLS;
		if ( offs + (LINES-2) * COLS > fsize ) {
			offs = fsize + COLS-1 - (LINES-2) * COLS;
			offs -= offs % COLS;
			if ( offs < 0 ) offs = 0;
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_UP:
		offs -= COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_NPAGE:
		offs += (LINES-2) * COLS;
		if ( offs + (LINES-2) * COLS > fsize ) {
			offs = fsize + COLS-1 - (LINES-2) * COLS;
			offs -= offs % COLS;
			if ( offs < 0 ) offs = 0;
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		offs -= (LINES-2) * COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_END:
		offs = fsize + COLS-1 - (LINES-2) * COLS;
		offs -= offs % COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = DISASM_MODE;
		goto leave_mode;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			int srch_from;
	
			srch_from = forw? patoffs+1: patoffs-1; 
			patoffs = do_search( srch_from, spat, lpat, forw );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( offs, &spat, &lpat, forw ) ) {
			patoffs = do_search( offs, spat, lpat, forw );
search_done:
			if ( patoffs != -1 ) {
				if ( patoffs < offs ||
				     patoffs > offs + (LINES-2)*COLS ) {
					offs = patoffs;
					offs -= offs % COLS;
				}
			}
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;
	}
  }

leave_mode:
  pop_keyhints();
}

void enter_offset( int& offs ) {
  int i, c, req_offs;
  char buf[9], *p;

  push_keyhints();
  clear_keyhints();

  move( 1, 0 );
  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = 0; i < 8; i++ ) addch( ACS_CKBOARD );

  move ( 1, 0 );
  p = buf;
  for (;;) {
  	c = getch();
	if ( c == (c & 0xff) && isxdigit(c) && p < buf + 8 ) {
		c = toupper(c);
		addch( c );
		*p++ = c;
	}
	else if ( c == KEY_BACKSPACE ) {
		p--;
		mvaddch( 1, p - buf, ACS_CKBOARD );
		move( 1, p - buf );
	}
	else if ( c == '\n' ) {
		*p = 0;
		break;
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
	else if ( c == 033 )
		goto cleanup;
  }

  req_offs = strtol( buf, 0, 16 );
  if ( req_offs >= 0 && req_offs < fsize ) offs = req_offs;

cleanup:
  pop_keyhints();
}

void interactive_asm( int offs ) {
  bool quit_requested = false;
  int xpos, ypos, c, minx, miny, maxy, maxx, i;
  char textbuf[50], *p;
  WINDOW *w;

  push_keyhints();
  clear_keyhints();
  set_keyhint( 10, "Exit" );
  refresh();

  w = newwin( 15, 51, (LINES-15)/2, (COLS-51)/2 );
  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Interactive Assembler" );

  minx = 18;
  miny = 1;
  getmaxyx( w, maxy, maxx );
  maxx -= 2;
  maxy -= 2;

  xpos = minx;
  ypos = miny;
  wmove( w, ypos, xpos );

  while ( !quit_requested ) {
  	c = wgetch( w );
	switch ( c ) {
	case KEY_DOWN:
		if ( ypos < maxy ) {
			ypos++;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_UP:
		if ( ypos > miny ) {
			ypos--;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_RIGHT:
		if ( xpos < maxx ) {
			xpos++;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_LEFT:
		if ( xpos > minx ) {
			xpos--;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_HOME:
		xpos = minx;
		wmove( w, ypos, xpos );
		break;
	case KEY_END:
		xpos = maxx;
		wmove( w, ypos, xpos );
		break;
	case '\n':
	case KEY_ENTER:
		if ( ypos < maxy ) {
			xpos = minx;
			ypos++;
			wmove( w, ypos, xpos );
		}
		break;
	case KEY_BACKSPACE:
		if ( xpos > minx ) {
			xpos--;
			mvwaddch( w, ypos, xpos, ' ' );
			wmove( w, ypos, xpos );
		}
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;
	default:
		if ( c == (c & 0xff) && c != 0x9b && isprint( c )  )
			mvwaddch( w, ypos, xpos, c );
		else
			mvwaddch( w, ypos, xpos, '.' );

		mvwinnstr( w, ypos, minx, textbuf, maxx - minx );

		p = textbuf;
		while ( *p == ' ' ) p++;
		memmove( textbuf, p, strlen(p) + 1 );
		p = textbuf + strlen( textbuf ) - 1;
		while ( p > textbuf && *p == ' ' ) {
			*p = 0;
			p--;
		}

		wmove( w, ypos, 1 );

		if ( do_asm_str( offs, textbuf ) ) {
			unsigned char *asmbuf;
			int asmlen;

// fix, pass current_ip instead of offs
			asmbuf = get_asm_data( asmlen );
			for ( i = 0; i < asmlen && i < 8; i++ )
				wprintw( w, "%02X", asmbuf[i] );
		}
		else i = 0;

		for ( ; i < 8; i++ )
			waddstr( w, "  " );

		xpos++;
		if ( xpos > maxx ) {
			if ( ypos < maxy ) {
				xpos = minx;
				ypos++;
			}
			else
				xpos = maxx;
		}
		wmove( w, ypos, xpos );
	}
  }

leave_mode:
  delwin( w );
  pop_keyhints();
}

int main( int argc, char **argv ) {
  int offs;

  if ( argc < 2 ) {
  	puts( "Syntax: ./disasm <filename>" );
	return 1;
  }

  if ( !strncmp( argv[1], "gdb:", 4 ) ) {
  	gdb_connection = true;

  	if ( gdb_connect( argv[1] + 4 ) < 0 ) {
		perror( "gdb_connect" );
		return 1;
	}

	if ( gdb_readmem( (unsigned char*)buf, 0, sizeof buf ) < 0 )
		fsize = 0;
	else
		fsize = 0x7fffffff;
  }
  else {
	fd = open( argv[1], O_RDONLY );
	if ( fd == -1 ) {
	  	perror( "disasm open" );
		return 1;
	}
	read( fd, buf, sizeof buf );
	fsize = lseek( fd, 0, SEEK_END );
  }

  buf_offs = 0;

  if ( load_le_objtbl() )
  	view_objtbl = view_le_objtbl;
  else if ( load_pe_objtbl() )
  	view_objtbl = view_pe_objtbl;
  else if ( load_elf_objtbl() )
  	view_objtbl = view_elf_objtbl;

  switch ( elf_arch ) {
  case EM_386:      x86_mode    = true; break;
  case EM_X86_64:   x86_64_mode = true; break;
  case EM_ARM:      arm_mode    = true; break;
  case EM_OPENRISC: break;
  default:          x86_mode    = true;
  }

  setlocale( LC_CTYPE, "" );

  initscr();
  start_color();
  cbreak();
  noecho();
  disasm_init( argv[1] );
  keypad( stdscr, true );

  quit_requested = false;
  disasm_mode = DISASM_MODE;
  offs = 0;
  disasm_mode_init();

  while ( !quit_requested ) {
  	if ( disasm_mode == DISASM_MODE )
		disasm_loop( offs );
	else if ( disasm_mode == HEX_MODE )
		hex_loop( offs );
	else
		text_loop( offs );
  }

do_exit: 
  erase();
  refresh();
  endwin();
}
